@@ -1,12 +1,23 @@
+# -*- coding: utf-8 -*-
+
 # Copyright 2013 Canonical Ltd. This software is licensed under
 # the GNU Affero General Public License version 3 (see the file
 # LICENSE).
+
+from __future__ import unicode_literals
+
+try:
+    str = unicode
+except NameError:
+    pass  # Forward compatibility with Py3k
+
 import json
 import unittest
 
 from datetime import datetime
 
 from mock import (
+    ANY,
     MagicMock,
     patch,
 )
@@ -22,6 +33,9 @@ from ssoclient.v2.http import (
 from ssoclient.v2.client import V2ApiClient, datetime_from_string
 
 
+TEST_ENDPOINT = 'http://foo.com/'
+
+
 class DateTimeFromStringTestCase(unittest.TestCase):
     """Test for the datetime parser."""
 
@@ -62,7 +76,7 @@ class ApiSessionTestCase(unittest.TestCase):
     def test_api_session_post(self, mock_request):
         mock_request.return_value = MagicMock(status_code=200)
 
-        api = ApiSession('http://foo.com')
+        api = ApiSession(TEST_ENDPOINT)
         api.post('/foo', data=dict(x=1))
         mock_request.assert_called_one_with(
             'POST',
@@ -74,7 +88,7 @@ class ApiSessionTestCase(unittest.TestCase):
     def test_api_session_get(self, mock_request):
         mock_request.return_value = MagicMock(status_code=200)
 
-        api = ApiSession('http://foo.com')
+        api = ApiSession(TEST_ENDPOINT)
         api.get('/foo', params=dict(x=1))
         mock_request.request.assert_called_one_with(
             'POST',
@@ -87,7 +101,7 @@ def mock_response(status_code=200, content=None, json_dump=True):
     response = Response()
     response.status_code = status_code
     if content is not None and json_dump:
-        content = json.dumps(content)
+        content = str(json.dumps(content)).encode('utf-8')
     response._content = content
     return response
 
@@ -100,7 +114,7 @@ class ProcessResponseTestCase(unittest.TestCase):
         return response
 
     def test_error_code_raises_correct_exception(self):
-        for code, exc in ERRORS.iteritems():
+        for code, exc in ERRORS.items():
             response = mock_response(exc.status_code, dict(code=code))
             with self.assertRaises(exc):
                 self.do_test(response)
@@ -145,7 +159,7 @@ class V2ClientApiTestCase(unittest.TestCase):
 
     def setUp(self):
         super(V2ClientApiTestCase, self).setUp()
-        self.client = V2ApiClient('http://foo.com')
+        self.client = V2ApiClient(TEST_ENDPOINT)
 
         p = patch('ssoclient.v2.http.requests.Session.request')
         self.mock_request = p.start()
@@ -236,7 +250,7 @@ class V2ClientApiTestCase(unittest.TestCase):
             credentials['consumer_key'], credentials['consumer_secret'],
             credentials['token_key'], credentials['token_secret'],
         )
-        self.assertTrue(all(isinstance(val, unicode) for
+        self.assertTrue(all(isinstance(val, str) for
                             val in self.mock_oauth.call_args[0]))
 
 
@@ -251,7 +265,7 @@ class RegisterV2ClientApiTestCase(V2ClientApiTestCase):
 
         self.mock_request.return_value = response
         with self.assertRaises(ExceptionClass) as ctx:
-            self.client.login(email='blah')
+            self.client.register(email='blah')
 
         if status_code >= 500:
             self.assertIn('some error message', str(ctx.exception))
@@ -383,8 +397,14 @@ class PasswordResetV2ClientApiTestCase(V2ClientApiTestCase):
         self.mock_request.assert_called_once_with(
             'POST', 'http://foo.com/tokens/password',
             headers={'Content-Type': 'application/json'},
-            data='{"token": null, "email": "%s"}' % self.email,
+            data=ANY
         )
+        # Load the data string into a dictionary so the order of the keys
+        # doesn't matter.
+        _, kwargs = self.mock_request.call_args
+        self.assertEqual(
+            {'token': None, 'email': self.email},
+            json.loads(kwargs.get('data')))
 
     def test_request_password_reset_without_email(self):
         self.mock_request.return_value = mock_response(
@@ -406,11 +426,17 @@ class PasswordResetV2ClientApiTestCase(V2ClientApiTestCase):
             self.email, 'token1234')
 
         self.assertEqual(response, content)
-        self.assertEqual(self.mock_request.call_args, [
-            ('POST', 'http://foo.com/tokens/password'),
-            {'headers': {'Content-Type': 'application/json'},
-             'data': '{"token": "token1234", "email": "%s"}' % self.email},
-        ])
+        self.mock_request.assert_called_once_with(
+            'POST', 'http://foo.com/tokens/password',
+            headers={'Content-Type': 'application/json'},
+            data=ANY
+        )
+        # Load the data string into a dictionary so the order of the keys
+        # doesn't matter.
+        _, kwargs = self.mock_request.call_args
+        self.assertEqual(
+            {'token': 'token1234', 'email': self.email},
+            json.loads(kwargs.get('data')))
 
     def test_request_password_reset_for_suspended_account(self):
         self.mock_request.return_value = mock_response(
@@ -634,16 +660,112 @@ class TokensV2ClientApiTestCase(V2ClientApiTestCase):
 
 class ValidateRequestV2ClientApiTestCase(V2ClientApiTestCase):
 
+    http_url = 'http://example.com'
+
+    def assert_validate_request_called(self, **kwargs):
+        data = dict(http_url=self.http_url, http_method='GET')
+        data.update(kwargs)
+        self.mock_request.assert_called_once_with(
+            'POST', TEST_ENDPOINT + 'requests/validate',
+            headers={'Content-Type': 'application/json'},
+            data=json.dumps(data))
+
     def test_valid_request(self):
-        self.mock_request.return_value = mock_response(200, dict(is_valid=True))
+        self.mock_request.return_value = mock_response(
+            200, dict(is_valid=True))
         result = self.client.validate_request(
-            http_url='foo', http_method='GET', authorization='123456789')
+            http_url=self.http_url, http_method='GET',
+            authorization='123456789')
 
         self.assertEqual(result, {'is_valid': True})
+        self.assert_validate_request_called(authorization='123456789')
 
     def test_invalid_request(self):
-        self.mock_request.return_value = mock_response(200, dict(is_valid=False))
+        self.mock_request.return_value = mock_response(
+            200, dict(is_valid=False))
         result = self.client.validate_request(
-            http_url='foo', http_method='GET', authorization='123456789')
+            http_url=self.http_url, http_method='GET',
+            authorization='123456789')
 
         self.assertEqual(result, {'is_valid': False})
+        self.assert_validate_request_called(authorization='123456789')
+
+    def test_missing_authorization(self):
+        self.mock_request.return_value = mock_response(
+            200, dict(is_valid=True))
+        result = self.client.validate_request(
+            http_url=self.http_url, http_method='GET',
+            query_string='oauth_signature=12345678&oauth_token=something')
+
+        self.assertEqual(result, {'is_valid': True})
+        self.assert_validate_request_called(
+            query_string='oauth_signature=12345678&oauth_token=something')
+
+    def test_missing_query_string(self):
+        self.mock_request.return_value = mock_response(
+            200, dict(is_valid=True))
+        result = self.client.validate_request(
+            http_url=self.http_url, http_method='GET',
+            authorization='123456789')
+
+        self.assertEqual(result, {'is_valid': True})
+        self.assert_validate_request_called(authorization='123456789')
+
+    def test_both_defined(self):
+        self.mock_request.return_value = mock_response(
+            200, dict(is_valid=True))
+        result = self.client.validate_request(
+            http_url=self.http_url, http_method='GET',
+            authorization='123456789',
+            query_string='oauth_signature=12345678&oauth_token=something')
+
+        self.assertEqual(result, {'is_valid': True})
+        self.assert_validate_request_called(
+            authorization='123456789',
+            query_string='oauth_signature=12345678&oauth_token=something')
+
+    def test_none_defined(self):
+        with self.assertRaises(errors.ClientError) as ctx:
+            self.client.validate_request(
+                http_url=self.http_url, http_method='GET')
+
+        self.assertIn(
+            'empty QUERY_STRING and no HTTP_AUTHORIZATION set',
+            str(ctx.exception))
+        self.assertFalse(self.mock_request.called)
+
+    def test_unicode_url(self):
+        self.mock_request.return_value = mock_response(
+            200, dict(is_valid=True))
+
+        http_url = 'http://localhost/~/test/doc/dåc-id'
+        result = self.client.validate_request(
+            http_url=http_url, http_method='GET', authorization='something')
+
+        self.assertEqual(result, {'is_valid': True})
+        self.assert_validate_request_called(
+            http_url=http_url, authorization='something')
+
+    def test_non_ascii_url_utf8_encoded(self):
+        self.mock_request.return_value = mock_response(
+            200, dict(is_valid=True))
+
+        http_url = 'http://localhost/~/test/doc/dåc-id'
+        result = self.client.validate_request(
+            http_url=http_url.encode('utf-8'), http_method='GET',
+            authorization='something')
+
+        self.assertEqual(result, {'is_valid': True})
+        self.assert_validate_request_called(
+            http_url=http_url, authorization='something')
+
+    def test_non_ascii_url_not_utf8_encoded(self):
+        http_url = 'http://localhost/~/test/doc/dåc-id'.encode('latin-1')
+
+        with self.assertRaises(errors.ClientError) as ctx:
+            self.client.validate_request(
+                http_url=http_url, http_method='GET', authorization='foobar')
+
+        e = ctx.exception
+        self.assertIn(repr(http_url), str(e))
+        self.assertFalse(self.mock_request.called)
